from copy import deepcopy

from flask import request
from flask import current_app
from flask_sqlalchemy import BaseQuery
from sqlalchemy.ext.declarative import DeclarativeMeta

from flask_atomic.orm.base import DeclarativeBase
from sqlalchemy.exc import InvalidRequestError
from sqlalchemy.orm import load_only
from sqlalchemy import desc

from sqlalchemy_blender import serialize
from sqlalchemy_blender import relationships
from sqlalchemy_blender import columns
from sqlalchemy_blender import serialize
from sqlalchemy_blender.helpers import related
from sqlalchemy_blender.query.processor import QueryStringProcessor

from handyhttp import HTTPNotFound
from handyhttp import HTTPBadRequest


class QueryBuffer:
    ACCEPTED_CALLS = ['all', 'count']

    def __init__(self, model, queryargs=None, auto=True, session=None):
        self.session = current_app.extensions['sqlalchemy'].db.session
        if not queryargs:
            queryargs = QueryStringProcessor(dict(request.args))

        if isinstance(model, DeclarativeMeta):
            query = self.session.query(model)
            self.basequery: BaseQuery = query
            self._basequery: BaseQuery = query
        else:
            self.basequery: BaseQuery = model.query
            self._basequery: BaseQuery = model.query

        self.pagedquery = None
        self.queryargs: QueryStringProcessor = queryargs
        self.data: [list, DeclarativeBase] = list()
        self.fields: set = set()
        self.count: int = 0
        self.flags = ['Y', 'N']
        self.paginated = False
        self.model = model
        self.active()

        if auto:
            self.apply()

    @staticmethod
    def querystring():
        return QueryStringProcessor(dict(request.args))

    def active(self):
        if getattr(self.model, 'active', None):
            self.basequery = self.basequery.filter(self.model.active != 'D')
            self.basequery = self.basequery.filter(self.model.active != 'X')
        return self

    @property
    def query(self):
        return self.basequery

    def check_key(self, key: str) -> bool:
        _key = key.split('.')
        for item in relationships(self.model):
            try:
                if _key[0] in [i.name for i in getattr(self.model, item).prop.target.columns]:
                    return True
            except Exception:
                continue
        return _key[0] in relationships(self.model)

    def prepare_query(self, fields):
        if not fields:
            self.basequery = self.session.query(self.model)
            return self.basequery
        select = [getattr(self.model, _field) for _field in fields]
        query = self.session.query(*select)
        self.basequery = query
        return self.basequery

    def apply(self, pending=False, inactive=False):
        entity = self.basequery._entity_zero()
        base_fields = set(map(lambda x: x.key, entity.column_attrs))
        _relationships = set(map(lambda x: x.key, entity.relationships))
        keys = columns(self.model, strformat=True)

        self.fields = base_fields.difference(set(self.queryargs.exclusions))
        # Now detect whether we want relationships
        if self.queryargs.include:
            fields = set()
            for item in self.queryargs.include:
                if not {item}.issubset(base_fields):
                    raise HTTPBadRequest('{} is not a recognised attribute of this resource'.format(item))
                fields.add(item)
            self.fields = fields

        if self.queryargs.rels:
            if type(self.queryargs.rels) in [list, set]:
                self.fields.union(set(relationships(self.model)).difference(self.queryargs.rels))
            elif self.queryargs.rels:
                self.fields.union(_relationships)
                self.queryargs.rels = _relationships

        updated_rels = set()

        if type(self.queryargs.rels) in [list, set]:
            for item in self.queryargs.rels:
                if self.check_key(item):
                    pass
                updated_rels.add(item)
            self.queryargs.rels = updated_rels

        if not self.queryargs.sortkey or self.queryargs.sortkey == '':
            self.queryargs.sortkey = entity.primary_key[0].name

        self.prepare_query(self.queryargs.include)

        if self.queryargs.sortkey in keys:
            column = getattr(self.basequery._entity_zero().attrs, self.queryargs.sortkey)
            if self.queryargs.descending and self.queryargs.sortkey in columns(self.model, strformat=True):
                self.basequery = self.basequery.order_by(desc(column))
            else:
                self.basequery = self.basequery.order_by(column)

        filters = self.queryargs.filters
        self.basequery, filters = self.check_relationship_filtering(self.basequery, filters)

        setfilters = []

        if getattr(self.queryargs, 'filterin', None):
            for key, search in getattr(self.queryargs, 'filterin').items():
                setfilters.append(getattr(self.model, key).in_(search))
                del filters[key]
                # self.basequery = self.basequery.filter(getattr(self.model, key).in_(search))

        # filters = tuple(filter(lambda k: '.' not in k, filters.keys()))

        if 'active' in keys:
            if pending:
                self.flags.append('P')
                self.flags.append('C')
            if inactive:
                self.flags.append('N')
            self.flags.append('C')
            setfilters.append(getattr(self.model, 'active').in_(self.flags))
            # self.basequery = self.set_active_filter(self.basequery, self.flags)

        self.basequery = self.basequery.filter(*setfilters)

        try:
            self.basequery = self.basequery.filter_by(**filters)
        except InvalidRequestError as exc:
            raise HTTPBadRequest(str(exc))

        for key, value in self.queryargs.max:
            self.basequery = self.basequery.filter(getattr(self.model, key) <= value)

        for key, value in self.queryargs.min:
            self.basequery = self.basequery.filter(getattr(self.model, key) >= value)

        if self.queryargs.page:
            self.pagedquery = self.basequery.paginate(self.queryargs.page, self.queryargs.pagesize or 50, False)
            self.paginated = True
            return self

        default_field = getattr(self.model, self.model.__mapper__.primary_key[0].name)
        self.basequery = self.basequery.group_by(default_field)
        self.basequery = self.basequery.limit(self.queryargs.limit)
        # self.basequery = self.basequery.options(load_only(*self.fields))
        return self

    def set_active_filter(self, query, flags):
        new_query = query
        if 'active' in columns(self.model, strformat=True) and self.model.active:
            new_query = query.filter(self.model.active.in_(flags))
        return new_query

    def check_relationship_filtering(self, query, filters):
        _filters = filters.copy()
        for item in filters.keys():
            splitter = item.split('.')
            if len(splitter) == 2:
                attribute = getattr(self.model, splitter[0]).comparator.entity.class_
                query = query.join(attribute).filter(getattr(attribute, splitter[1]) == filters[item])
                del _filters[item]
        return query, _filters

    def all(self, counts=True):
        if self.paginated:
            self.data = self.pagedquery.items
            self.count = getattr(self.basequery, 'count')()
            return self
            # raise ValueError('Cannot run all on a paginated query object')
        self.data = self.basequery.all()

        if counts:
            for idx, item in enumerate(self.data):
                rels = related(item)
                counts = {}
                for rel in rels:
                    field = getattr(rel, item.__tablename__, None)
                    if not field:
                        continue
                    resp = rel.query.filter(field.expression.right == item.id).count()
                    counts[rel.__tablename__] = resp
                setattr(self.data[idx], '__counts__', counts)
        self.count = len(self.data)
        return self

    def get(self, lookup, field=None):
        if field:
            if isinstance(field, str):
                field = getattr(self.model, field)
            return self._basequery.filter(field == lookup).first()
        primary_key = self.model.__mapper__.primary_key
        if isinstance(primary_key, tuple):
            primary_key = primary_key[0]
        return self._basequery.filter(primary_key == lookup).first()

    def one(self, field, value):
        filter_expression = {field: value}
        self.prepare_query(self.queryargs.include)
        query = self.basequery.filter_by(**filter_expression)
        self.data = query.first()
        if not self.data:
            raise HTTPNotFound(f'{str(self.model.__dict__.get("__tablename__")).capitalize()} not found!')
        for item in set([str(i.key) for i in relationships(self.model)]).intersection(self.queryargs.include):
            _item = item
            if not isinstance(_item, str):
                _item = str(item).split('.').pop()
            if isinstance(getattr(self.data, _item), BaseQuery):
                query = getattr(self.data, _item)
                related = QueryBuffer(query, deepcopy(self.queryargs), query._entity_zero().class_)
                related.basequery = related.set_active_filter(related.basequery, self.flags)
                related.fields = related.model.keys().difference(set(self.queryargs.exclusions))
                related_query_data = related.apply().all()
                setattr(self.data, '__i__' + item, related_query_data.json(autodata=False))
        return self




import uuid

from sqlalchemy.dialects.mysql.base import MSBinary
from sqlalchemy import types


class UUID(types.TypeDecorator):
    @property
    def python_type(self):
        pass

    def process_literal_param(self, value, dialect):
        pass

    impl = MSBinary

    def __init__(self):
        self.impl.length = 16
        types.TypeDecorator.__init__(self,length=self.impl.length)

    def process_bind_param(self,value,dialect=None):
        if value and isinstance(value,uuid.UUID):
            return value.bytes
        elif value and not isinstance(value,uuid.UUID):
            raise ValueError(f'{value} is not a valid uuid.UUID')
        else:
            return None

    def process_result_value(self,value,dialect=None):
        if value:
            return uuid.UUID(bytes=value)
        else:
            return None

    def is_mutable(self):
        return False



from copy import deepcopy

from flask import request
from flask_sqlalchemy import BaseQuery
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.exc import InvalidRequestError
from sqlalchemy import desc
from handyhttp.exceptions import HTTPNotFound
from handyhttp.exceptions import HTTPBadRequest

from .database import db
from .processor import QueryStringProcessor
from .helpers import relationships
from .helpers import columns


class QueryBuffer:
    ACCEPTED_CALLS = ['all', 'count']

    def __init__(self, model, queryargs=None, auto=True, session=None):
        self.session = session or db.session
        if not queryargs:
            queryargs = QueryStringProcessor(dict(request.args))

        if isinstance(model, DeclarativeMeta):
            query = self.session.query(model)
            self.basequery: BaseQuery = query
            self._basequery: BaseQuery = query
        else:
            self.basequery: BaseQuery = model.query
            self._basequery: BaseQuery = model.query

        self.pagedquery = None
        self.queryargs: QueryStringProcessor = queryargs
        self.data= list()
        self.fields: set = set()
        self.count: int = 0
        self.flags = ['Y']
        self.paginated = False
        self.model = model

        if auto:
            self.apply()

    def json(self, autodata=True, private=False):
        resp = []
        if not autodata:
            for item in self.data:
                resp.append(item.serialize(self.fields, rels=self.queryargs.rels, private=private))
            return resp

        if isinstance(self.data, list):
            for item in self.data:
                resp.append(item.serialize(self.fields, rels=self.queryargs.rels, private=private))
        else:
            return dict(data=self.data.serialize(self.fields, rels=self.queryargs.rels, private=private))
        return dict(data=resp, count=self.count)

    def check_key(self, key: str) -> bool:
        _key = key.split('.')
        for item in relationships(self.model):
            try:
                if _key[0] in [i.name for i in getattr(self.model, item).prop.target.columns]:
                    return True
            except Exception:
                continue
        return _key[0] in relationships(self.model)

    def prepare_query(self, fields):
        if not fields:
            self.basequery = self.session.query(self.model)
            return self.basequery
        select = [getattr(self.model, _field) for _field in fields]
        query = self.session.query(*select)
        self.basequery = query
        return self.basequery

    def apply(self, pending=False, inactive=False):
        entity = self.basequery._entity_zero()
        base_fields = set(map(lambda x: x.key, entity.column_attrs))
        _relationships = set(map(lambda x: x.key, entity.relationships))
        keys = columns(self.model, strformat=True)

        self.fields = base_fields.difference(set(self.queryargs.exclusions))
        # Now detect whether we want relationships
        if self.queryargs.include:
            fields = set()
            for item in self.queryargs.include:
                if not {item}.issubset(base_fields):
                    raise HTTPBadRequest('{} is not a recognised attribute of this resource'.format(item))
                fields.add(item)
            self.fields = fields

        if self.queryargs.rels:
            if type(self.queryargs.rels) in [list, set]:
                self.fields.union(set(relationships(self.model)).difference(self.queryargs.rels))
            elif self.queryargs.rels:
                self.fields.union(_relationships)
                self.queryargs.rels = _relationships

        updated_rels = set()

        if type(self.queryargs.rels) in [list, set]:
            for item in self.queryargs.rels:
                if self.check_key(item):
                    pass
                updated_rels.add(item)
            self.queryargs.rels = updated_rels

        if not self.queryargs.sortkey or self.queryargs.sortkey == '':
            self.queryargs.sortkey = entity.primary_key[0].name

        self.prepare_query(self.queryargs.include)

        if self.queryargs.sortkey in keys:
            column = getattr(self.basequery._entity_zero().attrs, self.queryargs.sortkey)
            if self.queryargs.descending and self.queryargs.sortkey in columns(self.model, strformat=True):
                self.basequery = self.basequery.order_by(desc(column))
            else:
                self.basequery = self.basequery.order_by(column)

        filters = self.queryargs.filters
        self.basequery, filters = self.check_relationship_filtering(self.basequery, filters)

        setfilters = []

        if getattr(self.queryargs, 'filterin', None):
            for key, search in getattr(self.queryargs, 'filterin').items():
                setfilters.append(getattr(self.model, key).in_(search))
                del filters[key]
                # self.basequery = self.basequery.filter(getattr(self.model, key).in_(search))

        # filters = tuple(filter(lambda k: '.' not in k, filters.keys()))

        if 'active' in keys:
            if pending:
                self.flags.append('P')
                self.flags.append('C')
            if inactive:
                self.flags.append('N')
            self.flags.append('C')
            setfilters.append(getattr(self.model, 'active').in_(self.flags))
            # self.basequery = self.set_active_filter(self.basequery, self.flags)

        self.basequery = self.basequery.filter(*setfilters)

        try:
            self.basequery = self.basequery.filter_by(**filters)
        except InvalidRequestError as exc:
            raise HTTPBadRequest(str(exc))

        for key, value in self.queryargs.max:
            self.basequery = self.basequery.filter(getattr(self.model, key) <= value)

        if self.queryargs.page:
            self.pagedquery = self.basequery.paginate(self.queryargs.page, self.queryargs.pagesize or 50, False)
            self.paginated = True
            return self

        default_field = getattr(self.model, self.model.__mapper__.primary_key[0].name)
        self.basequery = self.basequery.group_by(default_field)
        self.basequery = self.basequery.limit(self.queryargs.limit)
        # self.basequery = self.basequery.options(load_only(*self.fields))
        return self

    def set_active_filter(self, query, flags):
        new_query = query
        if 'active' in columns(self.model, strformat=True) and self.model.active:
            new_query = query.filter(self.model.active.in_(flags))
        return new_query

    def check_relationship_filtering(self, query, filters):
        _filters = filters.copy()
        for item in filters.keys():
            splitter = item.split('.')
            if len(splitter) == 2:
                attribute = getattr(self.model, splitter[0]).comparator.entity.class_
                query = query.join(attribute).filter(getattr(attribute, splitter[1]) == filters[item])
                del _filters[item]
        return query, _filters

    def all(self):
        if self.paginated:
            self.data = self.pagedquery.items
            self.count = getattr(self.basequery, 'count')()
            return self
            # raise ValueError('Cannot run all on a paginated query object')
        self.data = self.basequery.all()
        self.count = len(self.data)
        return self

    def one(self, field, value):
        filter_expression = {field: value}
        self.prepare_query(self.queryargs.include)
        query = self.basequery.filter_by(**filter_expression)
        self.data = query.first()
        if not self.data:
            raise HTTPNotFound(f'{str(self.model.__dict__.get("__tablename__"))} not found!')
        for item in set([str(i.key) for i in relationships(self.model)]).intersection(self.queryargs.include):
            _item = item
            if not isinstance(_item, str):
                _item = str(item).split('.').pop()
            if isinstance(getattr(self.data, _item), BaseQuery):
                query = getattr(self.data, _item)
                related = QueryBuffer(query, deepcopy(self.queryargs), query._entity_zero().class_)
                related.basequery = related.set_active_filter(related.basequery, self.flags)
                related.fields = related.model.keys().difference(set(self.queryargs.exclusions))
                related_query_data = related.apply().all()
                setattr(self.data, '__i__' + item, related_query_data.json(autodata=False))
        return self




from datetime import datetime

from flask import current_app
from sqlalchemy.exc import IntegrityError
from sqlalchemy.exc import ProgrammingError
from sqlalchemy_blender.helpers import columns
from sqlalchemy_blender.helpers import relationships

from handyhttp.exceptions import HTTPConflict
from handyhttp.exceptions import HTTPNotFound
from handyhttp.exceptions import HTTPNotProcessable
from handyhttp.exceptions import HTTPBadRequest


def getsession():
    db = current_app.extensions['sqlalchemy'].db
    return db.session


class ModelDAO:

    def __init__(self, model=None, autoupdate=True, session=None, *args, **kwargs):
        self.model = model
        self.autoupdate = autoupdate
        self.__session = session

    def session(self):
        if not self.__session:
            db = current_app.extensions['sqlalchemy'].db
            return db.session
        return self.session()

    def one(self, value, key=None):
        if not key:
            key = self.model.__mapper__.primary_key[0].name
        filter_expression = {key: value}
        query = self.model.query.filter_by(**filter_expression)
        return query.first()

    def validate_arguments(self, payload):
        valid_fields = dir(self.model)
        invalid_fields = []

        for item in payload:
            if not getattr(self.model, item, None):
                invalid_fields.append(item)
        if invalid_fields:
            err = f'<{invalid_fields}> not accepted fields'
            raise HTTPBadRequest(msg=err)

        return True

    def commit(self, operation, instance):
        session = self.session()
        try:
            session = current_app.extensions['sqlalchemy'].db.session
            getattr(session, operation)(instance)
            session.commit()
            return instance
        except IntegrityError as exc:
            session.rollback()
            if exc.orig.args[0] == 1452:
                raise HTTPNotProcessable()
            raise HTTPConflict(str(exc.orig.args[1]))
        except ProgrammingError as exc:
            if exc.orig.args[0] == 1064:
                raise HTTPNotProcessable()
        except Exception as exc:
            session = current_app.extensions['sqlalchemy'].db.session
            try:
                local_object = session.merge(session)
                session.add(local_object)
                session.commit()

                session = session.object_session()
                getattr(session, operation)(instance)
                session.commit()
            except Exception:
                session.rollback()
                session.flush()
                # session.closed()

    def delete(self, instance):
        if not instance:
            raise HTTPNotFound()
        return self.commit('delete', instance)

    def save(self, instance):
        if self.autoupdate:
            if getattr(instance, 'updated', None):
                setattr(instance, 'updated', datetime.now())
        self.commit('add', instance)
        return instance

    def create(self, payload, json=False, **kwargs):
        self.validate_arguments(payload)
        _payload = {}
        _keys = set([key for key in payload.keys() if not key.startswith('__')])
        for item in set([i.key for i in relationships(self.model)]).intersection(_keys):
            actual = getattr(relationships(self.model), item).primaryjoin.right.name
            _payload[actual] = payload.get(item)
            _payload[item] = payload.get(item)
            del payload[item]

        instance = self.model(**payload)
        if 'user' in kwargs.keys() and 'creator_id' in columns(self.model, strformat=True):
            instance.creator_id = kwargs.get('user')

        self.update(instance, _payload)
        return self.save(instance)

    def remove(self, instance, item):
        getattr(instance, item).remove(item)

    def update(self, instance, payload, **kwargs):
        fields = columns(instance, strformat=True)
        rels = relationships(instance)
        for attr, value in payload.items():
            if attr != 'id' and attr in fields:
                if str(getattr(instance.__mapper__.columns, attr).type) == 'DATETIME' and isinstance(value, bool):
                    value = datetime.now()
                setattr(instance, attr, value)

        for mtm in set([i.key for i in rels]).intersection(set(payload.keys())):
            model = getattr(instance, mtm)

            if not model and not isinstance(model, list):
                _instance = getattr(instance.__mapper__.relationships, mtm).entity.entity
                _lookup = mtm

                if isinstance(payload.get(mtm), dict):
                    _lookup = payload.get(mtm).get('id', None)

                elem = self.session().query(_instance).get(_lookup)
                if not elem:
                    continue
                setattr(instance, f'{mtm}_id', elem.id)
                continue

            if not isinstance(model, list):
                _instance = getattr(instance.__mapper__.relationships, mtm).entity.entity
                _lookup = payload.get(mtm).get('id', None)
                elem = self.session().query(_instance).get(_lookup)
                setattr(instance, mtm, elem)
                continue

            # current = set(map(lambda rel: getattr(rel, rel[0].__mapper__.primary_key[0].name), model))
            current = set(map(lambda item: getattr(item, 'id'), model))
            candidates = set(map(lambda item: item.get('id'), payload[mtm]))
            for addition in candidates.difference(current):
                association = self.session().query(
                    getattr(instance.__mapper__.relationships, mtm).entity
                ).get(addition)
                getattr(instance, mtm).append(association)

            for removal in current.difference(candidates):
                association = self.session().query(
                    getattr(instance.__mapper__.relationships, mtm).entity
                ).get(removal)
                getattr(instance, mtm).remove(association)

        self.save(instance)
        return instance

    def softdelete(self, instance, flag):
        instance.active = flag
        self.save(instance)



from typing import Optional
from datetime import datetime
from datetime import date

from sqlalchemy.orm.collections import InstrumentedList

from .database import db


def save(session):
    session.save()


def primarykey(model):
    return model.__mapper__.primary_key[0].name


def hidden_fields(element):
    return set([getattr(col, 'name', col) for col in getattr(element, '__hidden__', set())])


def update(self, commit=True, **kwargs):
    for attr, value in kwargs.items():
        if attr != 'id' and attr in self.fields():
            setattr(self, attr, value)

    for mtm in set(self.relationattrs()).intersection(set(kwargs.keys())):
        model = getattr(self, mtm)
        current = set(map(lambda rel: getattr(rel, rel.identify_primary_key()), model))
        candidates = set(map(lambda item: list(item.values()).pop(), kwargs[mtm]))
        for addition in candidates.difference(current):
            association = db.session.query(self.__mapper__.relationships.classgroups.entity).get(addition)
            getattr(self, mtm).append(association)

        for removal in current.difference(candidates):
            association = db.session.query(self.__mapper__.relationships.classgroups.entity).get(removal)
            getattr(self, mtm).remove(association)
    self.save()
    return self


def related(model, lookup=None):
    if isinstance(model, tuple):
        return list()

    if lookup:
        return getattr(model, lookup).entity.entity

    related_models = list()
    for i in model.__mapper__.relationships:
        related_models.append(i.entity.entity)
    return related_models


def relationships(model, fmt=False):
    rels = model.__mapper__.relationships
    if fmt:
        return set(map(lambda x: x.key, rels))
    return rels


def columns(model, strformat=False, relations=None):
    if not model:
        return None
    bound_columns = set(model.__mapper__.columns)
    if relations:
        return bound_columns.union(set([i.class_attribute for i in model.__mapper__.relationships]))
    if strformat:
        return [i.name for i in bound_columns]
    return bound_columns


def getschema(model):
    cols = set([i.name for i in columns(model)])
    cols = cols.difference([i.name for i in getattr(model, 'hidden', [])])
    schemamap = {
        'model': model.__tablename__,
        'fields': []
    }

    for item in cols:
        column = getattr(model, item)
        schemamap['fields'].append(dict(name=column.name, type=str(column.type)))

    return schemamap


def extract(element, fields=None, exclude: Optional[set] = None, **kwargs) -> Optional[dict]:
    resp = dict()
    if element is None:
        return None
    if exclude is None:
        exclude = set()

    if fields is None:
        fields = columns(element, strformat=True)

    for column in set(fields or []).difference(set(exclude)).difference(hidden_fields(element)):
        if isinstance(getattr(element, column), datetime) or isinstance(getattr(element, column), date):
            resp[column] = str(getattr(element, column))
        else:
            resp[column] = getattr(element, column)

    for prop in getattr(element, '__properties__', set()):
        resp[prop.fget.__name__] = getattr(element, prop.fget.__name__)

    if getattr(element, '__counts__', None):
        resp['__counts__'] = getattr(element, '__counts__')

    if getattr(element, '__expose__', None):
        for entity in getattr(element, '__expose__'):
            if entity.key in exclude:
                continue
            if entity.uselist:
                resp[entity.key] = []
                for child in getattr(element, entity.key):
                    resp[entity.key].append(extract(child))
            else:
                resp[entity.key] = extract(getattr(element, entity.key, None))

    return resp


def process_relationship(data, exclude):
    resp = None
    if isinstance(data, list):
        resp = []
        for item in data:
            resp.append(extract(item, columns(item, strformat=True), exclude))
        return resp
    else:
        resp = extract(data, columns(data, strformat=True), exclude)
    return resp


def transform(data, fields=None, relations=None, **kwargs):
    if isinstance(data, list):
        resp = []
        for item in data:
            resp.append(extract(item))
        return resp
    return extract(data)


def iserialize(data, fields=None, rels=None, root=None, exclude=None, functions=None,
               **kwargs):
    """
    This utility function dynamically converts Alchemy model classes into a
    dict using introspective lookups. This saves on manually mapping each
    model and all the fields. However, exclusions should be noted. Such as
    passwords and protected properties.

    :param model: SQLAlchemy model
    :param data: query data
    :param functions:
    :param fields: More of a whitelist of fields to include (preferred way)
    :param rels: Whether or not to introspect to relationships
    :param exc: Fields to exclude from query result set
    :param root: Root model for processing relationships. This acts as a
    recursive sentinel to prevent infinite recursion due to selecting oneself
    as a related model, and then infinitely trying to traverse the roots
    own relationships, from itself over and over.
    :param exclude: Exclusion in set form. Currently in favour of exc param.

    Only remedy to this is also to use one way relationships. Avoiding any
    back referencing of models.

    :return: json data structure of model
    :rtype: dict
    """

    if functions is None:
        functions = {}
    if exclude is None:
        exclude = set()
    else:
        exclude = set(exclude)

    model = data
    if isinstance(data, list) and len(data):
        model = data[0]

    exclude = exclude.union(set(kwargs.get('exclusions', set())))
    exclude.update(map(lambda col: getattr(col, 'name', col), getattr(model, 'hidden', set())))
    exclude.update(map(lambda col: getattr(col, 'name', col), getattr(model, '__hidden__', set())))

    if not data:
        return []

    if rels is True:
        rels = relationships(model, True)

    if not fields:
        fields = set(columns(model, strformat=True))

    fields = fields.difference(exclude)

    def process(element):
        if getattr(element, '_fields', None):
            transformed = {}
            for idx, field in enumerate(element._fields):
                transformed[field] = element[idx]
            return transformed

        transformed = extract(element, fields, exclude, **kwargs)
        if functions:
            for key, value in functions.items():
                transformed[f'_{key}'] = value(getattr(element, key))
        # rels = set([i.key for i in element.__mapper__.relationships]).intersection(fields)

        for item in rels or []:
            if '.' in item:
                left, right = item.split('.')
                _ = process_relationship(getattr(element, left), exclude)

                for idx, i in enumerate(getattr(element, left)):
                    _[idx][right] = extract(getattr(i, right))
                transformed[left] = _
                continue

            rel = None
            if getattr(element, item, None):
                rel = process_relationship(getattr(element, item), exclude)
                transformed[item] = rel
        return transformed

    if root is None:
        root = model.__tablename__

    # Define our model properties here. Columns and Schema relationships
    if not isinstance(data, list):
        return process(data)
    resp = []
    for element in data:
        resp.append(process(element))
    return resp


def serialize(model, data, fields=None, exc: Optional[set] = None, rels=None, root=None, exclude=None, functions=None,
              **kwargs):
    """
    This utility function dynamically converts Alchemy model classes into a
    dict using introspective lookups. This saves on manually mapping each
    model and all the fields. However, exclusions should be noted. Such as
    passwords and protected properties.

    :param model: SQLAlchemy model
    :param data: query data
    :param functions:
    :param fields: More of a whitelist of fields to include (preferred way)
    :param rels: Whether or not to introspect to relationships
    :param exc: Fields to exclude from query result set
    :param root: Root model for processing relationships. This acts as a
    recursive sentinel to prevent infinite recursion due to selecting oneself
    as a related model, and then infinitely trying to traverse the roots
    own relationships, from itself over and over.
    :param exclude: Exclusion in set form. Currently in favour of exc param.

    Only remedy to this is also to use one way relationships. Avoiding any
    back referencing of models.

    :return: json data structure of model
    :rtype: dict
    """

    if functions is None:
        functions = {}
    if exclude is None:
        exclude = set()
    else:
        exclude = set(exclude)

    exclude = exclude.union(set(kwargs.get('exclusions', set())))

    if rels is True:
        rels = relationships(model, True)

    exclude.update(map(lambda col: getattr(col, 'name', col), getattr(model, 'hidden', set())))
    exclude.update(map(lambda col: getattr(col, 'name', col), getattr(model, '__hidden__', set())))

    if not fields:
        fields = set(columns(model, strformat=True))

    fields = fields.difference(exclude)

    def process(element):
        if getattr(element, '_fields', None):
            transformed = {}
            for idx, field in enumerate(element._fields):
                transformed[field] = element[idx]
            return transformed

        transformed = extract(element, fields, set(), **kwargs)
        if functions:
            for key, value in functions.items():
                transformed[f'_{key}'] = value(getattr(element, key))
        # rels = set([i.key for i in element.__mapper__.relationships]).intersection(fields)

        for item in rels or []:
            if '.' in item:
                left, right = item.split('.')
                _ = process_relationship(getattr(element, left), exclude)

                for idx, i in enumerate(getattr(element, left)):
                    _[idx][right] = extract(getattr(i, right))
                transformed[left] = _
                continue

            rel = None
            if getattr(element, item, None):
                rel = process_relationship(getattr(element, item), exclude)
            transformed[item] = rel
        return transformed

    if root is None:
        root = model.__tablename__

    # Define our model properties here. Columns and Schema relationships
    if not isinstance(data, list):
        return process(data)
    resp = []
    for element in data:
        resp.append(process(element))
    return resp



from sqlalchemy import Column
from sqlalchemy import String
from sqlalchemy import Integer
from sqlalchemy import DateTime
from sqlalchemy import func


class PrimaryKeyMixin:
    id = Column(Integer, primary_key=True)


class LabelMixin:
    label = Column(String(256), default='')


class CreationTimestampMixin:
    created = Column(DateTime, default=func.now())


class UpdateMixin:
    updated = Column(DateTime, default=func.now())


class ActiveMixin:
    active = Column(String(5), default='Y')


class StatusMixin:
    status = Column(String(5), default='Y')


class SuperMixin(PrimaryKeyMixin, CreationTimestampMixin, UpdateMixin, ActiveMixin):
    __abstract__ = True



QUERYSTRING_ARGUMENT_MAP = {
    'true': True,
    'false': False,
    'default': None,
    'null': None
}

QUERYSTRING_CONTROL_KEYS = [
    'exclude',
    'limit', 'page',
    'order_by', 'desc',
    'min', 'max', 'only',
    'count', 'pagesize',
    'page', 'only'
]


class QueryStringProcessor:

    def __init__(self, querystring):
        self.querystring = querystring
        self.exclusions = list()
        self.filters = dict()
        self.namefilters = tuple()
        self.sortkey = str()
        self.limit = 100
        self.rels = False
        self.min = list()
        self.max = list()
        self.include = set()
        self.counts = {}
        self.descending = False
        self.page = None
        self.pagesize = 50
        self.__process_querystring()

    def splitspecial(self, key, operand):
        unpacked = key.split(operand)
        return (unpacked.pop(0), unpacked.pop())

    def __process_querystring(self):
        if not self.querystring:
            return None

        filterkeys = filter(
            lambda i: i[0] != 'relationships' and i[0] not in QUERYSTRING_CONTROL_KEYS, self.querystring.items()
        )
        for key, value in filterkeys:
            if '>' in key:
                _key, _value = self.splitspecial(key, '>')
                self.min.append((_key, _value,))
            elif '<' in key:
                _key, _value = self.splitspecial(key, '<')
                self.max.append((_key, _value,))
            elif QUERYSTRING_ARGUMENT_MAP.get(value) is False:
                self.exclusions.append(key)
            else:
                # Then this value filter is enabled
                self.filters[key] = value

        for item, value in self.filters.items():
            if ',' in value:
                setattr(self, 'filterin', {item: self.filters.get(item).split(',')})

        rels = self.querystring.get('relationships', None)
        if rels and rels not in ['false', 'N', 'no', 'No', '0']:
            if QUERYSTRING_ARGUMENT_MAP.get(rels) is not None:
                self.rels = QUERYSTRING_ARGUMENT_MAP.get(rels)
            else:
                self.rels = rels.split(',')

        only = self.querystring.get('only', None)
        if only:
            fields = self.querystring.get('only')

        if self.querystring.get('only', None):
            for field in self.querystring.get('only').split(','):
                self.include.add(field)

        order = self.querystring.get('order_by', False)
        if order:
            self.sortkey = order

        limit = self.querystring.get('limit', False)
        if limit:
            self.limit = int(limit)

        descending = self.querystring.get('desc', False)
        if descending:
            self.descending = QUERYSTRING_ARGUMENT_MAP.get(descending, None)

        pagesize = self.querystring.get('pagesize', False)
        if pagesize:
            self.pagesize = int(pagesize)

        page = self.querystring.get('page', False)
        if page:
            self.page = int(page)

        counts = self.querystring.get('count', None)
        if counts:
            for field in counts.split(','):
                # Like a partial function, call it on the relations when ready
                self.counts[field] = len

        gt = self.querystring.get('gt', False)
        if gt:
            self.gt = gt



from sqlalchemy_blender.helpers import serialize
from sqlalchemy_blender.helpers import iserialize
from sqlalchemy_blender.helpers import relationships
from sqlalchemy_blender.helpers import columns
from sqlalchemy_blender import helpers
from sqlalchemy_blender.database import db

from sqlalchemy_blender.query.buffer import QueryBuffer

__all__ = [relationships, columns, serialize, iserialize, helpers, QueryBuffer, db]

import flask_sqlalchemy

'''Created by Isaac Martin 2017. Licensed insofar as it can be according to the standard terms of the MIT license: 
https://en.wikipedia.org/wiki/MIT_License. The author accepts no liability for consequences resulting from the use of 
this software. '''


class SQLAlchemy(flask_sqlalchemy.SQLAlchemy):
    def __init__(
            self,
            app=None,
            use_native_unicode=True,
            session_options=None,
            metadata=None,
            engine_options=None,
            query_class=flask_sqlalchemy.BaseQuery,
            model_class=flask_sqlalchemy.Model
    ):

        self.use_native_unicode = use_native_unicode
        self.Query = query_class
        self.session = self.create_scoped_session(session_options)
        self.Model = self.make_declarative_base(model_class, metadata)
        self._engine_lock = flask_sqlalchemy.Lock()
        self.app = app
        self._engine_options = engine_options or {}
        flask_sqlalchemy._include_sqlalchemy(self, query_class)
        self.external_bases = []

        if app is not None:
            self.init_app(app)

    def get_tables_for_bind(self, bind=None):
        """
        Returns a list of all tables relevant for a bind.
        """

        result = []
        for Base in self.bases:
            for table in flask_sqlalchemy.itervalues(Base.metadata.tables):
                if table.info.get('bind_key') == bind:
                    result.append(table)

        return result

    def get_binds(self, app=None):
        """
        Returns a dictionary with a table->engine mapping.
        This is suitable for use of sessionmaker(binds=db.get_binds(app)).
        """

        app = self.get_app(app)
        binds = [None] + list(app.config.get('SQLALCHEMY_BINDS') or ())
        retval = {}
        for bind in binds:
            engine = self.get_engine(app, bind)
            tables = self.get_tables_for_bind(bind)
            retval.update(dict((table, engine) for table in tables))
        return retval

    @property
    def bases(self):
        return [self.Model] + self.external_bases

    def register_base(self, Base):
        """
        Register an external raw SQLAlchemy declarative base.
        Allows usage of the base with our session management and
        adds convenience query property using self.Query by default.
        """

        self.external_bases.append(Base)
        for c in Base._decl_class_registry.values():
            if isinstance(c, type):
                if not hasattr(c, 'query') and not hasattr(c, 'query_class'):
                    c.query_class = self.Query
                if not hasattr(c, 'query'):
                    c.query = flask_sqlalchemy._QueryProperty(self)

                    # for name in dir(c):
                    #     attr = getattr(c, name)
                    #     if type(attr) == orm.attributes.InstrumentedAttribute:
                    #         if hasattr(attr.prop, 'query_class'):
                    #             attr.prop.query_class = self.Query

                    # if hasattr(c , 'rel_dynamic'):
                    #     c.rel_dynamic.prop.query_class = self.Query



from setuptools import find_packages
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='SQLAlchemyBlender',
    version='0.1.3',
    description='Helpers package for SQL-Alchemy',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='python library utilities',
    url='https://github.com/kmjbyrne/sqlalchemy-blender',
    author='Keith Byrne',
    author_email='keithmbyrne@gmail.com',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'markdown',
        'twine'
    ],
    test_suite='nose.collector',
    tests_require=['nose', 'nose-cover3'],
    include_package_data=True,
    zip_safe=False
)



from flask import current_app
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm.attributes import InstrumentedAttribute

from sqlalchemy_blender.helpers import iserialize
from flask_atomic.database import db
# from flask_atomic.orm.database import db
from flask_atomic.orm.operators import commitsession


def session():
    db = current_app.extensions.get('sqlalchemy', None).db
    if not db:
        raise RuntimeError('A DB instance is required')
    return db.session


class JSONMixin:

    def json(self):
        return iserialize(self)


class CoreMixin(object):
    __abstract__ = True


    @declared_attr
    def __tablename__(self):
        return self.__name__.lower()

    @classmethod
    def normalise(cls, field):
        """
        Checks whether filter or field key is an InstumentedAttribute and returns
        a usable string instead. InstrumentedAttributes are not compatible with
        queries. Therefore, they must be transformed to correct types.

        :param field: The field we need to check
        :return: Transformed field
        """

        if isinstance(field, InstrumentedAttribute):
            return field.name
        return field

    @classmethod
    def fields(cls, inc=None, exc=None):
        if inc is None:
            inc = []
        if exc is None:
            exc = []

        normalised_fields = []
        for field in list(key for key in cls.keys() if
                          key not in [cls.normalise(e) for e in exc]):
            normalised_fields.append(cls.normalise(field))
        return normalised_fields

    @classmethod
    def create(cls, **payload):
        instance = cls()
        return instance.update(commit=True, **payload)

    def delete(self):
        x = current_app
        session().delete(self)
        commitsession()

    def save(self, commit=True):
        session().add(self)
        x = current_app
        if commit:
            commitsession()
        return self

    def update(self, commit=True, **kwargs):
        for attr, value in kwargs.items():
            if attr != 'id' and attr in self.fields():
                setattr(self, attr, value)

        for mtm in set(self.relationattrs()).intersection(set(kwargs.keys())):
            model = getattr(self, mtm)
            current = set(map(lambda rel: getattr(rel, rel.identify_primary_key()), model))
            candidates = set(map(lambda item: list(item.values()).pop(), kwargs[mtm]))
            for addition in candidates.difference(current):
                association = session().query(self.__mapper__.relationships.classgroups.entity).get(addition)
                getattr(self, mtm).append(association)

            for removal in current.difference(candidates):
                association = session().query(self.__mapper__.relationships.classgroups.entity).get(removal)
                getattr(self, mtm).remove(association)
        self.save()
        return self

    def commit(self):
        return commitsession()
